"use strict";
// Copyright (c) Martin Renou
// Distributed under the terms of the Modified BSD License.
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
const version_1 = require("./version");
class CanvasModel extends base_1.DOMWidgetModel {
    defaults() {
        return Object.assign({}, super.defaults(), { _model_name: CanvasModel.model_name, _model_module: CanvasModel.model_module, _model_module_version: CanvasModel.model_module_version, _view_name: CanvasModel.view_name, _view_module: CanvasModel.view_module, _view_module_version: CanvasModel.view_module_version, size: [700, 500] });
    }
    initialize(attributes, options) {
        super.initialize(attributes, options);
        this.canvas = document.createElement('canvas');
        this.ctx = this.canvas.getContext('2d');
        this.resizeCanvas();
        this.on('change:size', this.resizeCanvas.bind(this));
        this.on('msg:custom', this.onCommand.bind(this));
    }
    onCommand(command, buffers) {
        this.processCommand(command, buffers);
        this.forEachView((view) => {
            view.updateCanvas();
        });
    }
    processCommand(command, buffers) {
        if (command instanceof Array) {
            let remainingBuffers = buffers;
            for (const subcommand of command) {
                let subbuffers = [];
                if (subcommand.n_buffers) {
                    subbuffers = remainingBuffers.slice(0, subcommand.n_buffers);
                    remainingBuffers = remainingBuffers.slice(subcommand.n_buffers);
                }
                this.processCommand(subcommand, subbuffers);
            }
            return;
        }
        switch (command.name) {
            case 'putImageData':
                this.putImageData(command.args, buffers);
                break;
            case 'set':
                this.setAttr(command.attr, command.value);
                break;
            case 'clear':
                this.clearCanvas();
                break;
            default:
                this.executeCommand(command.name, command.args);
                break;
        }
    }
    putImageData(args, buffers) {
        const [bufferMetadata, dx, dy] = args;
        const data = new Uint8ClampedArray(buffers[0].buffer);
        const imageData = new ImageData(data, bufferMetadata.shape[1], bufferMetadata.shape[0]);
        this.ctx.putImageData(imageData, dx, dy);
    }
    setAttr(attr, value) {
        this.ctx[attr] = value;
    }
    clearCanvas() {
        this.forEachView((view) => {
            view.clear();
        });
        this.ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
    }
    executeCommand(name, args) {
        this.ctx[name](...args);
    }
    forEachView(callback) {
        for (const view_id in this.views) {
            this.views[view_id].then((view) => {
                callback(view);
            });
        }
    }
    resizeCanvas() {
        const size = this.get('size');
        this.canvas.setAttribute('width', size[0]);
        this.canvas.setAttribute('height', size[1]);
    }
}
CanvasModel.serializers = Object.assign({}, base_1.DOMWidgetModel.serializers);
CanvasModel.model_name = 'CanvasModel';
CanvasModel.model_module = version_1.MODULE_NAME;
CanvasModel.model_module_version = version_1.MODULE_VERSION;
CanvasModel.view_name = 'CanvasView';
CanvasModel.view_module = version_1.MODULE_NAME;
CanvasModel.view_module_version = version_1.MODULE_VERSION;
exports.CanvasModel = CanvasModel;
class CanvasView extends base_1.DOMWidgetView {
    render() {
        this.canvas = document.createElement('canvas');
        this.el.appendChild(this.canvas);
        this.ctx = this.canvas.getContext('2d');
        this.resizeCanvas();
        this.model.on('change:size', this.resizeCanvas.bind(this));
        this.updateCanvas();
    }
    clear() {
        this.ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
    }
    updateCanvas() {
        this.ctx.drawImage(this.model.canvas, 0, 0);
    }
    resizeCanvas() {
        const size = this.model.get('size');
        this.canvas.setAttribute('width', size[0]);
        this.canvas.setAttribute('height', size[1]);
    }
}
exports.CanvasView = CanvasView;
class MultiCanvasModel extends base_1.DOMWidgetModel {
    defaults() {
        return Object.assign({}, super.defaults(), { _model_name: MultiCanvasModel.model_name, _model_module: MultiCanvasModel.model_module, _model_module_version: MultiCanvasModel.model_module_version, _view_name: MultiCanvasModel.view_name, _view_module: MultiCanvasModel.view_module, _view_module_version: MultiCanvasModel.view_module_version, _canvases: [] });
    }
}
MultiCanvasModel.serializers = Object.assign({}, base_1.DOMWidgetModel.serializers, { _canvases: { deserialize: base_1.unpack_models } });
MultiCanvasModel.model_name = 'MultiCanvasModel';
MultiCanvasModel.model_module = version_1.MODULE_NAME;
MultiCanvasModel.model_module_version = version_1.MODULE_VERSION;
MultiCanvasModel.view_name = 'MultiCanvasView';
MultiCanvasModel.view_module = version_1.MODULE_NAME;
MultiCanvasModel.view_module_version = version_1.MODULE_VERSION;
exports.MultiCanvasModel = MultiCanvasModel;
class MultiCanvasView extends base_1.DOMWidgetView {
    render() {
        this.container = document.createElement('div');
        this.container.style.position = 'relative';
        this.el.appendChild(this.container);
        this.canvas_views = new base_1.ViewList(this.createCanvasView, this.removeCanvasView, this);
        this.updateCanvasViews();
        this.model.on('change:_canvases', this.updateCanvasViews.bind(this));
    }
    updateCanvasViews() {
        this.canvas_views.update(this.model.get('_canvases'));
    }
    createCanvasView(canvasModel, index) {
        // The following ts-ignore is needed due to ipywidgets's implementation
        // @ts-ignore
        return this.create_child_view(canvasModel).then((canvasView) => {
            canvasView.el.style.zIndex = index;
            canvasView.el.style.position = 'absolute';
            this.container.appendChild(canvasView.el);
            return canvasView;
        });
    }
    removeCanvasView(canvasView) {
        this.container.removeChild(canvasView.el);
    }
}
exports.MultiCanvasView = MultiCanvasView;
//# sourceMappingURL=widget.js.map